from invisibleroads_macros_configuration import set_default
from invisibleroads_posts.routines.configuration import set_attribute

from .constants import RECORD_ID_LENGTH
from .models import (
    define_get_database_session,
    get_database_engine,
    get_transaction_manager_session)
from .variables import RECORDS_REGISTRY


def includeme(config):
    configure_settings(config)
    configure_models(config)


def configure_settings(config):
    settings = config.get_settings()
    set_attribute(
        settings, RECORDS_REGISTRY, 'id.length', RECORD_ID_LENGTH, int)
    set_default(
        settings, 'sqlalchemy.url',
        'sqlite:///%s/database.sqlite' % settings.get('data.folder', '.'))


def configure_models(config):
    settings = config.get_settings()
    settings['tm.manager_hook'] = 'pyramid_tm.explicit_manager'
    config.include('pyramid_tm')
    config.include('pyramid_retry')
    database_engine = get_database_engine(settings)
    get_database_session = define_get_database_session(database_engine)
    config.add_request_method(
        lambda r: get_transaction_manager_session(get_database_session, r.tm),
        'db', reify=True)



NAMING_CONVENTION = {
    'ix': 'ix_%(column_0_label)s',
    'uq': 'uq_%(table_name)s_%(column_0_name)s',
    'ck': 'ck_%(table_name)s_%(constraint_name)s',
    'fk': 'fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s',
    'pk': 'pk_%(table_name)s',
}
RECORD_ID_LENGTH = 8
RECORD_RETRY_COUNT = 3



from invisibleroads.scripts import ConfigurableScript
from invisibleroads_posts.routines.configuration import (
    load_bootstrapped_settings)

from .models import Base, get_database_engine


class InitializeRecordsScript(ConfigurableScript):

    priority = 20

    def run(self, args, argv):
        settings = load_bootstrapped_settings(args.configuration_path)
        if 'sqlalchemy.url' not in settings:
            return
        database_engine = get_database_engine(settings)
        Base.metadata.create_all(database_engine)



from invisibleroads_macros_log import get_log
from pyramid.view import exception_view_config
from sqlalchemy.exc import OperationalError


@exception_view_config(OperationalError)
def handle_database_operational_error(context, request):
    response = request.response
    response.status_int = 500
    L.error(DATABASE_ERROR_MESSAGE % context)
    return response


L = get_log(__name__)


DATABASE_ERROR_MESSAGE = """%s

Did you run the initialization script?

invisibleroads initialize development.ini
invisibleroads initialize production.ini"""



from datetime import datetime
from invisibleroads_macros_log import get_timestamp
from invisibleroads_macros_security import make_random_string
from invisibleroads_posts.views import get_value
from pyramid.httpexceptions import HTTPNotFound
from sqlalchemy import Column, engine_from_config
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from sqlalchemy.schema import MetaData
from sqlalchemy.types import DateTime, String
from zope.sqlalchemy import register as register_transaction_listener

from .constants import NAMING_CONVENTION, RECORD_ID_LENGTH, RECORD_RETRY_COUNT
from .exceptions import InvisibleRoadsRecordsError
from .variables import RECORDS_REGISTRY


class RecordMixin(object):

    id = Column(String, primary_key=True)
    id_length = RECORD_ID_LENGTH

    @classmethod
    def make_unique_record(Class, database, retry_count=RECORD_RETRY_COUNT):
        count = 0
        id_length = Class.id_length
        while count < retry_count:
            record = Class(id=make_random_string(id_length))
            database.add(record)
            try:
                database.flush()
            except IntegrityError:
                database.rollback()
            else:
                break
            count += 1
        else:
            raise InvisibleRoadsRecordsError(
                f'could not get unique {Class.__tablename__}.id')
        return record

    @classmethod
    def get_from(Class, request, record_id=None):
        key = Class.__tablename__ + 'Id'
        if record_id is None:
            record_id = get_value(request, key)
        database = request.database
        record = Class.get(database, record_id)
        if not record:
            raise HTTPNotFound({key: 'bad'})
        return record

    @classmethod
    def get(Class, database, record_id):
        if record_id is None:
            return
        return database.query(Class).get(record_id)

    def __repr__(self):
        return f'<{self.__class__.__name__}(id={self.id})>'


class CreationMixin(object):

    creation_datetime = Column(DateTime, default=datetime.utcnow)

    @property
    def creation_timestamp(self):
        return get_timestamp(self.creation_datetime)


class ModificationMixin(object):

    modification_datetime = Column(DateTime, default=datetime.utcnow)

    @property
    def modification_timestamp(self):
        return get_timestamp(self.modification_datetime)


def get_database_engine(settings, prefix='sqlalchemy.'):
    engine = engine_from_config(settings, prefix)
    for DatabaseExtension in settings.get('database.extensions', []):
        DatabaseExtension(settings, prefix).configure(engine)
    return engine


def define_get_database_session(database_engine):
    get_database_session = sessionmaker()
    get_database_session.configure(bind=database_engine)
    return get_database_session


def get_transaction_manager_session(get_database_session, transaction_manager):
    database_session = get_database_session()
    register_transaction_listener(
        database_session, transaction_manager=transaction_manager)
    return database_session


metadata = Base = declarative_base(
    class_registry=RECORDS_REGISTRY,
    metadata=MetaData(naming_convention=NAMING_CONVENTION))



from os.path import abspath, dirname, join
from setuptools import find_packages, setup


ENTRY_POINTS = '''
[invisibleroads]
initialize = invisibleroads_records.scripts:InitializeRecordsScript
'''
APPLICATION_CLASSIFIERS = [
    'Programming Language :: Python',
    'Framework :: Pyramid',
    'Topic :: Internet :: WWW/HTTP',
    'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
    'License :: OSI Approved :: MIT License',
]
APPLICATION_REQUIREMENTS = [
    # architecture
    'invisibleroads >= 0.3.3',
    'invisibleroads-posts >= 0.7.16.1',
    # web
    'pyramid',
    # database
    # 'alembic',
    'pyramid-retry',
    'pyramid-tm',
    'sqlalchemy',
    'transaction',
    'zope.sqlalchemy',
    # test
    'pytest',
    # shortcut
    'invisibleroads-macros-configuration >= 1.0.6',
    'invisibleroads-macros-log >= 1.0.3',
    'invisibleroads-macros-security >= 1.0.1',
]
TEST_REQUIREMENTS = [
    'pytest-cov',
]
FOLDER = dirname(abspath(__file__))
DESCRIPTION = '\n\n'.join(open(join(FOLDER, x)).read().strip() for x in [
    'README.md', 'CHANGES.md'])


setup(
    name='invisibleroads-records',
    version='0.5.9.2',
    description='Web application database defaults',
    long_description=DESCRIPTION,
    long_description_content_type='text/markdown',
    classifiers=APPLICATION_CLASSIFIERS,
    author='Roy Hyunjin Han',
    author_email='rhh@crosscompute.com',
    url='https://github.com/invisibleroads/invisibleroads-records',
    keywords='web wsgi bfg pylons pyramid invisibleroads',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=True,
    extras_require={'test': TEST_REQUIREMENTS},
    install_requires=APPLICATION_REQUIREMENTS,
    entry_points=ENTRY_POINTS)



import transaction
from pyramid import testing
from pytest import fixture

from .models import (
    Base,
    define_get_database_session,
    get_database_engine,
    get_transaction_manager_session)


@fixture
def records_request(records_config, database, data_folder):
    request = testing.DummyRequest(
        db=database,
        data_folder=data_folder)
    request.json_body = {}
    yield request


@fixture
def records_config(records_settings):
    config = testing.setUp(settings=records_settings)
    config.include('invisibleroads_posts')
    config.include('invisibleroads_records')
    yield config
    testing.tearDown()


@fixture
def records_settings(posts_settings):
    records_settings = posts_settings
    records_settings['sqlalchemy.url'] = 'sqlite://'
    yield records_settings


@fixture
def database(records_settings, database_extensions):
    database_engine = get_database_engine(records_settings)
    for Extension in database_extensions:
        Extension(records_settings).configure(database_engine)
    Base.metadata.create_all(database_engine)
    get_database_session = define_get_database_session(database_engine)
    database_session = get_transaction_manager_session(
        get_database_session, transaction.manager)
    yield database_session
    transaction.abort()
    Base.metadata.drop_all(database_engine)


@fixture
def database_extensions():
    return []
